/** \file CahnHilliardNavierStokes.h */

#ifndef CAHN_HILLIARD_NAVIER_STOKES_H
#define CAHN_HILLIARD_NAVIER_STOKES_H

// coupling structures
#include "CouplingIterationInterface.h"
#include "CouplingTimeInterface.h"
#include "CouplingProblemStat.h"

// structures for local refinement
#include "Refinement.h"
#include "MeshFunction_Level.h"

#include "POperators.h"

using namespace AMDiS;

/**
  * \ingroup Problem 
  *
  * \brief
  */
template<typename CH_Type = CahnHilliard,
	 typename NS_Type = NavierStokes_TaylorHood>
class CahnHilliardNavierStokes : public CouplingIterationInterface,
                                 public CouplingTimeInterface,
                                 public CouplingProblemStat
{
public:

  CahnHilliardNavierStokes(std::string name_, CH_Type *chProb_, NS_Type *nsProb_)
  : CouplingProblemStat(name_),
    chProb(chProb_),
    nsProb(nsProb_),
    refFunction(NULL),
    refinement(NULL),
    sigma(0.0),
    surfaceTension(0.0)
  {
    dow = Global::getGeo(WORLD);

    Parameters::get(name + "->sigma", sigma);
    surfaceTension = sigma/chProb->getEpsilon();
  }


  ~CahnHilliardNavierStokes() {
    if (refFunction != NULL)
      delete refFunction;
    if (refinement != NULL)
      delete refinement;
  }


  /**
   * Add the problems to the iterationInterface, timeInterface and couplingProblemStat.
   * As a consequence all problem can be initialized one after another and in the
   * adaption loop they are solved in rotation.
   * At the end the problems are filled with operators and coupling operators as well as
   * boundary conditions are added.
   *
   * In the adaption loop the problems are solved the same order as they are added to the
   * iterationInterface in this method. This order can be changed manually in the oneIteration
   * method.
   **/
  void initialize(AdaptInfo *adaptInfo)
  {
    for (size_t i = 0; i < chProb->getNumProblems(); i++)
      addProblem(chProb->getProblem(i));
    for (size_t i = 0; i < nsProb->getNumProblems(); i++)
      addProblem(nsProb->getProblem(i));
    
    addIterationInterface(chProb);
    addIterationInterface(nsProb);

    addTimeInterface(chProb);
    addTimeInterface(nsProb);

    CouplingProblemStat::initialize(INIT_ALL);
    dim = getMesh()->getDim();

    fillCouplingOperators();
    // fillOperators and fillBoundaryConditions for chProb and nsProb
    nsProb->initTimeInterface();
    chProb->initTimeInterface();
//     fillCouplingBoundaryConditions();
  }


  /**
   * In this method the operators responsible for the coupling of all problems are defined
   * and added to the problem. The access to the problems is through the baseBroblem->getProblem()
   * method. All baseProblem provide FiniteElemSpaces and solution vectors.
   **/
  void fillCouplingOperators()
  { FUNCNAME("CahnHilliardNavierStokes::fillCouplingOperators()");
    MSG("CahnHilliardNavierStokes::fillCouplingOperators()");

    for (size_t i = 0; i < dow; i++) {
      // < nu * d_i(c) , theta >
      Operator *opNuGradC = new Operator(nsProb->getFeSpace(i), chProb->getFeSpace(0));
      opNuGradC->addTerm(new VecAndPartialDerivative_ZOT(chProb->getSolution()->getDOFVector(1), chProb->getSolution()->getDOFVector(0), i));
      nsProb->getProblem(0)->addVectorOperator(opNuGradC, i, &surfaceTension, &surfaceTension);
    }

    // < v * grad(c) , theta >
    Operator *opVGradC = new Operator(chProb->getFeSpace(0), chProb->getFeSpace(0));
    opVGradC->addTerm(new WorldVector_FOT(nsProb->getVelocity(), -1.0), GRD_PSI);
    chProb->getProblem()->addMatrixOperator(opVGradC, 0, 0);
  }
  

  /**
   * Solves the initial problems.Before this is done the mesh is refined by the Refinement-class
   * This provides local refinement functions for phaseField-refinement or signed-distance-refinement.
   * See MeshFunction_Level.h for some details.
   **/ 
  void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    // meshFunction for klocal refinement around the interface of the phasefield-DOFVector
    refFunction = new PhaseFieldRefinement(chProb->getMesh());

    if (chProb->getDoubleWellType() == 0) {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	chProb->getSolution()->getDOFVector(0)); // phaseField-DOFVector in [0,1]
    } else {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	new PhaseDOFView<double>(chProb->getSolution()->getDOFVector(0))); // phaseField-DOFVector in [-1,1]
    }

    // initial refinement
    refinement->refine(0);

    for (int i = 0; i < 3; i++) {
      chProb->solveInitialProblem(adaptInfo); 	// update phaseField-DOFVector
      refinement->refine((i < 4 ? 4 : 10));	// do k steps of local refinement
    }

    // solve all initial problems of the problems added to the CouplingTimeInterface
    CouplingTimeInterface::solveInitialProblem(adaptInfo);
  }


  /**
   * Called at the end of each timestep.
   * If the phase-field has changed the mesh is updated by the refinement-method
   **/
  void closeTimestep(AdaptInfo *adaptInfo)
  {
    CouplingTimeInterface::closeTimestep(adaptInfo);

    refinement->refine(2);
  }

protected:

  CH_Type *chProb;	// CahnHilliard baseProblem
  NS_Type *nsProb;	// Navier-Stokes baseProblem

  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;

  unsigned dim;		// dimension of the mesh
  unsigned dow;		// dimension of the world

  double sigma;		// coupling parameter to calculate the surface tension
  double surfaceTension;// := sigma/epsilon
};

#endif // CAHN_HILLIARD_NAVIER_STOKES_H
